import collections
import tensorflow as tf
import tensorflow.keras.applications as ka
from tensorflow.keras.layers import Dense, Dropout, Flatten, Conv2D, MaxPooling2D, Input, Lambda
from tensorflow.keras.models import Model, Sequential
from miso.layers.cyclic import *

# Input range is [0,1], not [0,255] as expected by keras
# So all prepro are adjusted to this.
from miso.layers.msoftmax import MSoftMaxLayer


def tf_prepro(input_shape):
    # (x / 127.5) - 1
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.subtract(tf.multiply(y, 2), 1))
    return inputs, x


def torch_prepro(input_shape):
    # (x / 255 - mean) / std_dev
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.subtract(y, tf.reshape(tf.constant([0.485, 0.456, 0.406]), [1, 1, 1, 3])))(inputs)
    x = Lambda(lambda y: tf.divide(y, tf.reshape(tf.constant([0.229, 0.224, 0.225]), [1, 1, 1, 3])))(x)
    return inputs, x


def default_prepro(input_shape):
    # Convert to BGR then x - mean
    inputs = Input(shape=input_shape)
    x = Lambda(lambda y: tf.reverse(y, axis=[-1]))(inputs)
    x = Lambda(lambda y: tf.subtract(tf.multiply(y, 255.0),
                                     tf.reshape(tf.constant([103.939, 116.779, 128.68]), [1, 1, 1, 3])))(x)
    return inputs, x


def head(cnn_type, input_shape):
    subtypes = cnn_type.split('_')
    model_type = subtypes[0]
    if 'cyclic' in subtypes:
        use_cyclic = True
    else:
        use_cyclic = False
    if 'gain' in subtypes:
        use_gain = True
    else:
        use_gain = False

    params = TRANSFER_LEARNING_PARAMS[model_type]
    inputs, x = params.prepro_func(input_shape)
    if use_cyclic:
        if use_gain:
            x = CyclicGainSlice12()(x)
        else:
            x = CyclicSlice4()(x)
    x = params.model_func(include_top=False, weights='imagenet', pooling='avg', input_shape=input_shape).call(x)
    if use_cyclic:
        if use_gain:
            x = CyclicDensePoolN(pool_op=tf.reduce_mean)(x)
        else:
            x = CyclicDensePool4(pool_op=tf.reduce_mean)(x)
    model = Model(inputs=inputs, outputs=x)
    for layer in model.layers:
        layer.trainable = False
    return model

# TODO make adjustable
def tail(num_classes, input_shape, dropout=(0.5, 0.5)):
    inp = Input(shape=input_shape)
    outp = Dropout(dropout[0])(inp)
    outp = Dense(512, activation='relu')(outp)
    # outp = Dropout(dropout[1])(outp)
    # outp = Dense(512, activation='relu')(outp)
    outp = Dense(num_classes, activation='softmax')(outp)
    return Model(inp, outp)


def tail_vector(num_classes, input_shape, dropout=(0.5, 0.5)):
    inp = Input(shape=input_shape)
    outp = Dropout(dropout[0])(inp)
    outp = Dense(512, activation='relu')(outp)
    # outp = Dropout(dropout[1])(outp)
    # outp = Dense(512, activation='relu')(outp)
    return Model(inp, outp)


TransferLearningParams = collections.namedtuple(
    'TransferLearningParams',
    ['model_func', 'prepro_func', 'default_input_shape']
)

# TODO Update for tensorflow 2
TRANSFER_LEARNING_PARAMS = {
    'xception': TransferLearningParams(ka.xception.Xception, tf_prepro, [299, 299, 3]),
    'vgg16': TransferLearningParams(ka.vgg16.VGG16, default_prepro, [224, 224, 3]),
    'vgg19': TransferLearningParams(ka.vgg19.VGG19, default_prepro, [224, 224, 3]),
    'resnet50': TransferLearningParams(ka.resnet50.ResNet50, default_prepro, [224, 224, 3]),
    # 'resnet101': TransferLearningParams(ka.xception.Xception, default_prepro, [224,224,3]),
    # 'resnet152': TransferLearningParams(ka.xception.Xception, default_prepro, [224,224,3]),
    # 'resnet50V2': TransferLearningParams(ka.xception.Xception, default_prepro, [224,224,3]),
    # 'resnet101V2': TransferLearningParams(ka.xception.Xception, default_prepro, [224,224,3]),
    # 'resnet152V2': TransferLearningParams(ka.xception.Xception, default_prepro, [224,224,3]),
    'inceptionV3': TransferLearningParams(ka.inception_v3.InceptionV3, tf_prepro, [299, 299, 3]),
    'inceptionresnetV2': TransferLearningParams(ka.inception_resnet_v2.InceptionResNetV2, tf_prepro, [299, 299, 3]),
    'mobilenet': TransferLearningParams(ka.mobilenet.MobileNet, tf_prepro, [224, 224, 3]),
    'mobilenetV2': TransferLearningParams(ka.mobilenet_v2.MobileNetV2, tf_prepro, [224, 224, 3]),
    'densenet121': TransferLearningParams(ka.densenet.DenseNet121, torch_prepro, [224, 224, 3]),
    'densenet169': TransferLearningParams(ka.densenet.DenseNet169, torch_prepro, [224, 224, 3]),
    'densenet201': TransferLearningParams(ka.densenet.DenseNet201, torch_prepro, [224, 224, 3]),
    'nasnetmobile': TransferLearningParams(ka.nasnet.NASNetMobile, tf_prepro, [224, 224, 3]),
    'nasnetlarge': TransferLearningParams(ka.nasnet.NASNetLarge, tf_prepro, [331, 331, 3])
}



"""
Creates and trains a generic network
"""
import warnings

warnings.simplefilter(action='ignore', category=FutureWarning)

import time
import datetime
from collections import OrderedDict
import tensorflow.keras.backend as K
from sklearn.manifold import TSNE

from miso.data.tf_generator import TFGenerator
from miso.data.training_dataset import TrainingDataset
from miso.stats.embedding import plot_embedding
from miso.stats.mislabelling import find_and_save_mislabelled
from miso.training.adaptive_learning_rate import AdaptiveLearningRateScheduler
from miso.training.training_result import TrainingResult
from miso.stats.confusion_matrix import *
from miso.stats.training import *
from miso.training.tf_augmentation import aug_all_fn
from miso.deploy.saving import freeze, convert_to_inference_mode, save_frozen_model_tf2, convert_to_inference_mode_tf2, load_from_xml
from miso.deploy.model_info import ModelInfo
from miso.models.factory import *

import matplotlib.pyplot as plt
import pandas as pd


def train_image_classification_model(tp: MisoParameters):
    tf_version = int(tf.__version__[0])

    # Hack to make RTX cards work
    if tf_version == 2:
        physical_devices = tf.config.list_physical_devices('GPU')
        for device in physical_devices:
            tf.config.experimental.set_memory_growth(device, True)
    else:
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        session = tf.Session(config=config)

    K.clear_session()

    # Clean the training parameters
    tp.sanitise()

    print("+---------------------------------------------------------------------------+")
    print("| MISO Particle Classification Library                                      |")
    print("+---------------------------------------------------------------------------+")
    print("| Stable version:                                                           |")
    print("| pip install miso2                                                         |")
    print("| Development version:                                                      |")
    print("| pip install git+http://www.github.com/microfossil/particle-classification |")
    print("+---------------------------------------------------------------------------+")
    print("Tensorflow version: {}".format(tf.__version__))
    print("-" * 80)
    print("Train information:")
    print("- name: {}".format(tp.name))
    print("- description: {}".format(tp.description))
    print("- CNN type: {}".format(tp.cnn.id))
    print("- image type: {}".format(tp.cnn.img_type))
    print("- image shape: {}".format(tp.cnn.img_shape))
    print()

    # Load data
    ds = TrainingDataset(tp.dataset.source,
                         tp.cnn.img_shape,
                         tp.cnn.img_type,
                         tp.dataset.min_count,
                         tp.dataset.map_others,
                         tp.dataset.val_split,
                         tp.dataset.random_seed,
                         tp.dataset.memmap_directory)
    ds.load()
    tp.dataset.num_classes = ds.num_classes

    # ------------------------------------------------------------------------------
    # Transfer learning
    # ------------------------------------------------------------------------------
    if tp.cnn.id.endswith('tl'):
        print('-' * 80)
        print("Transfer learning network training")
        start = time.time()

        # Generate head model and predict vectors
        model_head = generate_tl_head(tp.cnn.id, tp.cnn.img_shape)

        # Calculate vectors
        print("- calculating vectors")
        t = time.time()
        gen = ds.images.create_generator(32, shuffle=False, one_shot=True)
        if tf_version == 2:
            vectors = model_head.predict(gen.create())
        else:
            vectors = model_head.predict_generator(gen.create(), steps=len(gen))
        print("! {}s elapsed, ({}/{} vectors)".format(time.time() - t, len(vectors), len(ds.images.data)))

        # Clear session
        # K.clear_session()

        # Generate tail model and compile
        model_tail = generate_tl_tail(tp.dataset.num_classes, [vectors.shape[-1], ])
        model_tail.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])

        # Learning rate scheduler
        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=tp.training.alr_epochs,
                                               nb_drops=tp.training.alr_drops,
                                               verbose=1)
        print('-' * 80)
        print("Training")

        # Training generator
        train_gen = TFGenerator(vectors,
                                ds.cls_onehot,
                                ds.train_idx,
                                tp.training.batch_size,
                                shuffle=True,
                                one_shot=False,
                                oversample=tp.training.use_class_undersampling)

        # Validation generator
        if tf_version == 2:
            val_one_shot = True
        else:
            # One repeat for validation for TF1 otherwise we get end of dataset errors
            val_one_shot = False
        if tp.dataset.val_split > 0:
            val_gen = TFGenerator(vectors,
                                  ds.cls_onehot,
                                  ds.test_idx,
                                  tp.training.batch_size,
                                  shuffle=False,
                                  one_shot=val_one_shot)
        else:
            val_gen = None

        # Class weights (only if over sampling is not used)
        if tp.training.use_class_weights is True and tp.training.use_class_undersampling is False:
            class_weights = ds.class_weights
            print("- class weights: {}".format(class_weights))
            if tf_version == 2:
                class_weights = dict(enumerate(class_weights))
        else:
            class_weights = None
        if tp.training.use_class_undersampling:
            print("- class balancing using random over sampling")

        v = model_tail.predict(vectors[0:1])
        print(v[0, :10])

        # model_head.summary()

        # model = Model(inputs=model_head.input, outputs=model_tail(model_head.output))
        # vector_model = Model(model.inputs, model.get_layer(index=-2).get_output_at(0))
        # v = vector_model.predict(ds.images.data[0:1] / 255)
        # print(v[0, :10])
        # v = vector_model.predict(ds.images.data[0:1])
        # print(v[0, :10])

        # Train
        history = model_tail.fit_generator(train_gen.create(),
                                           steps_per_epoch=len(train_gen),
                                           validation_data=val_gen.create(),
                                           validation_steps=len(val_gen),
                                           epochs=tp.training.max_epochs,
                                           verbose=0,
                                           shuffle=False,
                                           max_queue_size=1,
                                           class_weight=class_weights,
                                           callbacks=[alr_cb])
        # Elapsed time
        end = time.time()
        training_time = end - start
        print("- training time: {}s".format(training_time))
        time.sleep(3)

        # Now we join the trained dense layers to the resnet model to create a model that accepts images as input
        # model_head = generate_tl_head(tp.cnn.id, tp.cnn.img_shape)
        model = combine_tl(model_head, model_tail)
        model.summary()

        # print(model.layers[-1])
        # print(model.layers[-2])
        #
        # vector_model = Model(inputs=model.inputs, outputs=model.layers[-2].get_output_at(1))
        # print(vector_model.layers[-1])
        # print(vector_model.layers[-2])
        # v = vector_model.predict(ds.images.data[0:1] / 255)
        # print(v[0, :10])
        # v = vector_model.predict(ds.images.data[0:1])
        # print(v[0, :10])

        # model = Model(inputs=model_head.input, outputs=model_tail(model_head.layers[-1].layers[-1].output))
        # model.summary()

        # print(model_tail.get_layer(index=-2).get_weights())
        #
        # vector_tensor = model_tail.get_layer(index=-2).get_output_at(0)
        # vector_model = Model(model_tail.inputs, vector_tensor)
        vector_model = generate_vector(model, tp.cnn.id)
        # v = vector_model.predict(vectors[0:1])
        # print(v[0, :10])
        #
        # vectors = model_head.predict(next(iter(gen.create())))
        # v = vector_model.predict(vectors[0:1])
        # print(v[0, :10])


    # ------------------------------------------------------------------------------
    # Full network train
    # ------------------------------------------------------------------------------
    else:
        print('-' * 80)
        print("Full network training")
        start = time.time()

        # Generate model
        model = generate(tp)

        # Augmentation
        if tp.augmentation.rotation is True:
            rotation_range = [0, 360]
        else:
            rotation_range = None
        if tp.training.use_augmentation is True:
            print("- using augmentation")
            augment_fn = aug_all_fn(rotation=rotation_range,
                                    gain=tp.augmentation.gain,
                                    gamma=tp.augmentation.gamma,
                                    zoom=tp.augmentation.zoom,
                                    gaussian_noise=tp.augmentation.gaussian_noise,
                                    bias=tp.augmentation.bias,
                                    random_crop=tp.augmentation.random_crop,
                                    divide=255)
        else:
            print("- NOT using augmentation")
            augment_fn = TFGenerator.map_fn_divide_255

        # Learning rate scheduler
        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=tp.training.alr_epochs,
                                               nb_drops=tp.training.alr_drops,
                                               verbose=1)

        # Training generator
        train_gen = ds.images.create_generator(tp.training.batch_size, map_fn=augment_fn)

        # Validation generator
        if tf_version == 2:
            val_one_shot = True
        else:
            # One repeat for validation for TF1 otherwise we get end of dataset errors
            val_one_shot = False
        if tp.dataset.val_split > 0:
            val_gen = ds.test_generator(tp.training.batch_size, shuffle=False, one_shot=val_one_shot)
        else:
            val_gen = None

        # Class weights
        if tp.training.use_class_weights is True and tp.training.use_class_undersampling is False:
            class_weights = ds.class_weights
            print("- class weights: {}".format(class_weights))
            if tf_version == 2:
                class_weights = dict(enumerate(class_weights))
        else:
            class_weights = None

        # Train the model
        history = model.fit_generator(train_gen.create(),
                                      steps_per_epoch=len(train_gen),
                                      validation_data=val_gen.create(),
                                      validation_steps=len(val_gen),
                                      epochs=tp.training.max_epochs,
                                      verbose=0,
                                      shuffle=False,
                                      max_queue_size=1,
                                      class_weight=class_weights,
                                      callbacks=[alr_cb])

        # Elapsed time
        end = time.time()
        training_time = end - start
        print()
        print("Total training time: {}s".format(training_time))
        time.sleep(3)

        # Vector model
        vector_model = generate_vector(model, tp.cnn.id)

    # ------------------------------------------------------------------------------
    # Results
    # ------------------------------------------------------------------------------
    print('-' * 80)
    print("Evaluating model")
    now = datetime.datetime.now()
    save_dir = os.path.join(tp.output.save_dir, "{0}_{1:%Y%m%d-%H%M%S}".format(tp.name, now))
    os.makedirs(save_dir, exist_ok=True)
    # Accuracy
    if tp.dataset.val_split > 0:
        y_true = ds.cls[ds.test_idx]
        gen = ds.test_generator(1, shuffle=False, one_shot=True)
        if tf_version == 2:
            y_prob = model.predict(gen.to_tfdataset())
        else:
            y_prob = model.predict_generator(gen.tf1_compat_generator(), len(gen))
        y_pred = y_prob.argmax(axis=1)
    else:
        y_true = np.asarray([])
        y_prob = np.asarray([])
        y_pred = np.asarray([])
    # Inference time
    print("- calculating inference time:", end='')
    max_count = np.min([128, len(ds.images.data)])
    inf_times = []
    for i in range(3):
        gen = ds.images.create_generator(32, idxs=np.arange(max_count), shuffle=False, one_shot=True)
        start = time.time()
        if tf_version == 2:
            model.predict(gen.to_tfdataset())
        else:
            model.predict_generator(gen.tf1_compat_generator(), len(gen))
        end = time.time()
        diff = (end - start) / max_count * 1000
        inf_times.append(diff)
        print(" {:.3f}ms".format(diff), end='')
    inference_time = np.median(inf_times)
    print(", median: {}".format(inference_time))
    # Store results
    # - fix to make key same for tensorflow 1 and 2
    if 'accuracy' in history.history:
        history.history['acc'] = history.history.pop('accuracy')
        history.history['val_acc'] = history.history.pop('val_accuracy')
    result = TrainingResult(tp,
                            history,
                            y_true,
                            y_pred,
                            y_prob,
                            ds.cls_labels,
                            training_time,
                            inference_time)
    print("- accuracy {:.2f}".format(result.accuracy * 100))
    print("- mean precision {:.2f}".format(result.mean_precision * 100))
    print("- mean recall {:.2f}".format(result.mean_recall * 100))
    # ------------------------------------------------------------------------------
    # Save results
    # ------------------------------------------------------------------------------
    if tp.description is None:
        tp.description = "{}: {} model trained on data from {} ({} images in {} classes).\n" \
                         "Accuracy: {:.1f} (P: {:.1f}, R: {:.1f}, F1 {:.1f})".format(
            tp.name,
            tp.cnn.id,
            tp.dataset.source,
            len(ds.filenames.filenames),
            len(ds.cls_labels),
            result.accuracy * 100,
            result.mean_precision * 100,
            result.mean_recall * 100,
            result.mean_f1_score * 100)

    # Create model info with all the parameters
    inputs = OrderedDict()
    inputs["image"] = model.inputs[0]
    outputs = OrderedDict()
    outputs["pred"] = model.outputs[0]
    outputs["vector"] = vector_model.outputs[0]
    info = ModelInfo(tp.name,
                     tp.description,
                     tp.cnn.id,
                     now,
                     "frozen_model.pb",
                     tp,
                     inputs,
                     outputs,
                     tp.dataset.source,
                     ds.cls_labels,
                     ds.filenames.cls_counts,
                     "rescale",
                     [255, 0, 1],
                     result.accuracy,
                     result.precision,
                     result.recall,
                     result.f1_score,
                     result.support,
                     result.epochs[-1],
                     training_time,
                     tp.dataset.val_split,
                     inference_time)
    # ------------------------------------------------------------------------------
    # Plots
    # ------------------------------------------------------------------------------
    # Plot the graphs
    # plot_model(model, to_file=os.path.join(save_dir, "model_plot.pdf"), show_shapes=True)
    print("-" * 80)
    print("Plotting")
    if tp.dataset.val_split > 0:
        print("- loss")
        plot_loss_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "loss_vs_epoch.pdf"))
        print("- accuracy")
        plot_accuracy_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "accuracy_vs_epoch.pdf"))
        print("- confusion matrix")
        plot_confusion_accuracy_matrix(y_true, y_pred, ds.cls_labels)
        plt.savefig(os.path.join(save_dir, "confusion_matrix.pdf"))
        plt.close('all')

    # Vectors for mislabeled and t-SNE
    print("- calculating vectors... ", end='')
    gen = ds.images.create_generator(1, shuffle=False, one_shot=True)
    if tf_version == 2:
        vectors = vector_model.predict(gen.create(), steps=len(gen))
    else:
        vectors = vector_model.predict_generator(gen.create(), steps=len(gen))
    print("{} total".format(len(vectors)))
    # print(vectors[0, :10])
    v = vector_model.predict(ds.images.data[0:1] / 255)
    # print(v[0, :10])
    if tp.output.save_mislabeled is True:
        print("- mislabeled")
        find_and_save_mislabelled(ds.images.data,
                                  vectors,
                                  ds.cls,
                                  ds.cls_labels,
                                  [os.path.basename(f) for f in ds.filenames.filenames],
                                  save_dir,
                                  11)

    # t-SNE
    print("- t-SNE (1024 vectors max)")
    idx = np.random.choice(range(len(vectors)), np.min((len(vectors), 1024)), replace=False)
    vec_subset = vectors[idx]
    X = TSNE(n_components=2).fit_transform(vec_subset)
    # counts = np.unique(ds.cls, return_counts=True)
    # print(counts)
    # counts = np.unique(ds.cls[idx], return_counts=True)
    # print(counts)
    plot_embedding(X, ds.cls[idx], ds.num_classes)
    plt.savefig(os.path.join(save_dir, "tsne.pdf"))
    # Legend
    cls_info = pd.DataFrame({"index": range(ds.num_classes), "label": ds.cls_labels})
    cls_info.to_csv(os.path.join(save_dir, "legend.csv"), sep=';')

    # ------------------------------------------------------------------------------
    # Save model (has to be last thing it seems)
    # ------------------------------------------------------------------------------
    print('-' * 80)
    print("Saving model")
    # Convert if necessary to fix TF batch normalisation issues

    # Freeze and save graph
    if tp.output.save_model is not None:
        if tf_version == 2:
            inference_model = convert_to_inference_mode_tf2(model, lambda: generate(tp))
            tf.saved_model.save(inference_model, os.path.join(os.path.join(save_dir, "model_keras")))
            frozen_func = save_frozen_model_tf2(inference_model, os.path.join(save_dir, "model"), "frozen_model.pb")
            info.inputs["image"] = frozen_func.inputs[0]
            info.outputs["pred"] = frozen_func.outputs[0]
        else:
            inference_model = convert_to_inference_mode(model, lambda: generate(tp))
            tf.saved_model.save(inference_model, os.path.join(os.path.join(save_dir, "model_keras")))
            freeze(inference_model, os.path.join(save_dir, "model"))

    # Save model info
    info.save(os.path.join(save_dir, "model", "network_info.xml"))

    # ------------------------------------------------------------------------------
    # Confirm model save
    # ------------------------------------------------------------------------------
    if tp.output.save_model is not None and tp.dataset.val_split > 0:
        print("-" * 80)
        print("Validate saved model")
        y_pred_old = y_pred
        y_true = ds.cls[ds.test_idx]
        gen = ds.test_generator(32, shuffle=False, one_shot=True)
        y_prob = []
        if tf_version == 2:
            model, img_size, cls_labels = load_from_xml(os.path.join(save_dir, "model", "network_info.xml"))
            for b in iter(gen.to_tfdataset()):
                y_prob.append(model(b[0]).numpy())
        else:
            session, input, output, img_size, cls_labels = load_from_xml(os.path.join(save_dir, "model", "network_info.xml"))
            iterator = iter(gen.tf1_compat_generator())
            for bi in range(len(gen)):
                b = next(iterator)
                y_p = session.run(output, feed_dict={input: b[0]})
                y_prob.append(y_p)
        y_prob = np.concatenate(y_prob, axis=0)
        y_pred = y_prob.argmax(axis=1)
        acc = accuracy_score(y_true, y_pred)
        p, r, f1, _ = precision_recall_fscore_support(y_true, y_pred)
        print("Saved model on test set: acc {:.2f}, prec {:.2f}, rec {:.2f}, f1 {:.2f}".format(acc, np.mean(p), np.mean(r), np.mean(f1)))
        acc = accuracy_score(y_pred_old, y_pred)
        if acc == 1.0:
            print("Overlap: {:.2f}% - PASSED".format(acc * 100))
        else:
            print("Overlap: {:.2f}% - FAILED".format(acc * 100))

    # ------------------------------------------------------------------------------
    # Clean up
    # ------------------------------------------------------------------------------
    print("- cleaning up")
    ds.release()
    print("- complete")
    print('-' * 80)
    print()
    return model, vector_model, ds, result



"""
Training parameters
"""
import json
import re
from collections import OrderedDict

from miso.models.transfer_learning import TRANSFER_LEARNING_PARAMS


class Parameters(object):
    def asdict(self):
        d = OrderedDict()
        for name in dir(self):
            v = getattr(self, name)
            if not name.startswith('__') and not callable(v):
                if isinstance(v, Parameters):
                    d[name] = v.asdict()
                else:
                    d[name] = v
        return d

    def to_json(self):
        d = self.asdict()
        return json.dumps(d)


class CNNParameters(Parameters):
    id = "base_cyclic"
    img_shape = [128, 128, 1]
    img_type = "greyscale"
    filters = 4
    blocks = None
    dense = None
    use_batch_norm = True
    global_pooling = None
    activation = "relu"
    use_asoftmax = False


class TrainingParameters(Parameters):
    batch_size = 64
    max_epochs = 10000
    alr_epochs = 10
    alr_drops = 4
    monitor_val_loss = False
    use_class_weights = True
    use_class_balancing = False
    use_augmentation = True


class DatasetParameters(Parameters):
    num_classes = None
    source = None
    min_count = 10
    val_split = 0.2
    map_others = False
    random_seed = 0
    memmap_directory = None


class AugmentationParameters(Parameters):
    rotation = [0, 360]
    gain = [0.8, 1, 1.2]
    gamma = [0.5, 1, 2]
    bias = None
    zoom = [0.9, 1, 1.1]
    gaussian_noise = None
    random_crop = None
    orig_img_shape = [256, 256, 3]


class OutputParameters(Parameters):
    output_dir = None
    save_model = True
    save_mislabeled = True


class MisoParameters(Parameters):
    name = ""
    description = ""
    cnn = CNNParameters()
    dataset = DatasetParameters()
    training = TrainingParameters()
    augmentation = AugmentationParameters()
    output = OutputParameters()

    def sanitise(self):
        if self.name == "":
            self.name = self.dataset.source + "_" + self.cnn.id
            self.name = re.sub('[^A-Za-z0-9]+', '-', self.name)
        if self.cnn.img_shape is None:
            if self.cnn.id.endswith("_tl"):
                shape = TRANSFER_LEARNING_PARAMS[self.cnn.id.split('_')[0]].default_input_shape
            else:
                if self.cnn.id.startswith("base_cyclic") or self.cnn.id.startswith("resnet_cyclic"):
                    shape = [128, 128, 3]
                else:
                    shape = [224, 224, 3]
                if self.cnn.img_type == 'rgb':
                    shape[2] = 3
                    self.augmentation.orig_img_shape[2] = 3
                else:
                    shape[2] = 1
                    self.augmentation.orig_img_shape[2] = 3
            self.cnn.img_shape = shape
        elif self.cnn.id.endswith("_tl"):
            self.cnn.img_shape[2] = 3
        if self.name == "":
            self.name = self.dataset.source.replace("http://", "").replace("https://", "").replace("/", "-").replace("\\", "-") + "_" + self.cnn.id + "_" + self.cnn.img_shape + "_" + self.cnn.img_type


def get_default_shape(cnn_type):
    if cnn_type.endswith("_tl"):
        return TRANSFER_LEARNING_PARAMS[cnn_type.split('_')[0]].default_input_shape
    else:
        return [224, 224, None]


if __name__ == "__main__":
    m = MisoParameters()
    print(m.asdict())
    print(m.to_json())










from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='miso2',
    version='2.3.18',
    description='Python scripts for training CNNs for particle classification',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ross Marchant',
    author_email='ross.g.marchant@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
    ],
    keywords='microfossil, cnn',
    python_requires='>=3.6, <4',
    packages=['miso', 'miso.data', 'miso.deploy', 'miso.layers', 'miso.models', 'miso.stats', 'miso.training', 'miso.utils'],
    install_requires=['image-classifiers>=1.0.0',
                      'lxml',
                      'matplotlib',
                      'numpy',
                      'pandas',
                      'Pillow',
                      'imagecodecs',
                      'scikit-image',
                      'scikit-learn',
                      'scipy',
                      'segmentation-models',
                      'dill',
                      'flask',
                      'tqdm',
                      'openpyxl',
                      'imblearn'],
    url='https://github.com/microfossil/particle-classification',
    license='MIT',
    project_urls={  # Optional
        'Source': 'https://github.com/microfossil/particle-classification',
        'Paper': 'https://jm.copernicus.org/articles/39/183/2020/',
    },
)



import argparse
from miso.training.trainer import train_image_classification_model
from miso.training.parameters import MisoParameters

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Train a CNN to classify images")
    parser.add_argument("-i", "--input", required=True, help="Directory of images, URL link to zipped directory of images, or ParticleTrieur project file")
    parser.add_argument("-o", "--output", required=True, help="Output directory to store training results")
    parser.add_argument("-t", "--type", required=True, help="Type of CNN: resnet_tl, resnet_cyclic_tl, base_cyclic, resnet_cyclic, resnet18, resnet50, vgg16, vgg19")

    parser.add_argument("-f", "--filters", type=int, default=4, help="Number of filters in the first convolutional block")

    parser.add_argument("--min_count", type=int, default=10, help="Minimum number of images in a class for it to be included")
    parser.add_argument("--map_others", action='store_true', help="Classes with not enough images will be put into 'others' class (so long as the total is also greater than min_count")
    args = parser.parse_args()

    tp = MisoParameters()
    tp.source = args.input
    tp.output_dir = args.output
    tp.cnn_type = args.type
    tp.filters = args.filters
    tp.min_count = args.min_count
    tp.map_others = args.map_others

    train_image_classification_model(tp)
    # print("This is the name of the script: ", sys.argv[0])
    # print("Number of arguments: ", len(sys.argv))
    # print("The arguments are: ", str(sys.argv))
    #
    # params = default_params()
    # print(params)
    # key = None
    # for arg in sys.argv[1:]:
    #     print(arg)
    #     if arg.startswith('--'):
    #         key = arg[2:]
    #         print(key)
    #     else:
    #         try:
    #             val = int(arg)
    #             print("int")
    #         except ValueError:
    #             try:
    #                 val = float(arg)
    #                 print("float")
    #             except:
    #                 val = arg
    #                 print("string")
    #         params[key] = val
    # print(params)
    # train_image_classification_model(params)

